//==========================================================================================================================
// AVR I2S audio output demo
// Version V1.00, 10.2.2012
// 
// (C) 2012, Stanislav Maslan, s.maslan@seznam.cz
// www: http://www.elektronika.kvalitne.cz/ATMEL/necoteorie/tutorial/AudioTest/I2Semulation/I2S_emulation.html
//==========================================================================================================================
// This simple program plays PCM wave files recieved via USART.
// It uses SPI to emulate I2S audio bus for audio DACs TDA1543, TDA1545 or similar.
// Data can be send by any terminal with support for CTS outgoing data flow control. 
// Stream end is detected automatically if no data are recieved for more than 750ms. 
// File format and data length is detected from RIFF container. A small disadvantage of 
// used RIFF parser is that "data" subchunk has to be located before 0x0078 offset (usually it is). 
//
// MCU:           ATmega644-20
// F_CPU:         16MHz typically
// DAC:           I2S without master clock (TDA1543 or similar)
//                Japaneese two's complement format (TDA1545)
// Serial format: 2Mbd, 1 stop, no parity, CTS handshake (can be modified)
// Wave format:   PCM wave, 8/16-bit, mono/stereo, sample rate up to 44.1kHz (48kHz is a bit tight for 2Mbd)
//
//
// FUSES: If crystal is used it should be cofigurated in full swing mode because crystal terminals are next to 
//        USART RXD terminal and since there is some 2MHz transmission there can be interference. In my case oscillator 
//        became very unstable during the transmission and data were currupted.
//==========================================================================================================================

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <stdlib.h>

#include "main.h"
#include "serial.h"
#include "sound.h"


//==========================================================================================================================
// Timeout functions
//==========================================================================================================================
// ocr1a ISR
volatile uint8_t to_cnt;
/*ISR(TIMER1_COMPA_vect,ISR_NOBLOCK)
{
	uint8_t cnt=to_cnt;
	if(cnt)
		to_cnt=--cnt;
}*/
ISR(TIMER1_COMPA_vect,ISR_NAKED)
{
	asm(
		"sei \n"
		"push R16 \n"
		"push R17 \n"
		"in R17,%0 \n"
		"lds R16,to_cnt \n"
		"tst R16 \n"
		"breq .+6 \n"
		"dec R16 \n"
		"sts to_cnt,R16 \n"
		"out %0,R17 \n"
		"pop R17 \n"
		"pop R16 \n"
		"reti"
		::"I" _SFR_IO_ADDR(SREG)
	);
}







//==========================================================================================================================
// Wave functions
//==========================================================================================================================
// globals
TWave wave;

// ------ parse RIFF container and configurate player ------
uint8_t riff_config(void)
{
	// get some data
	uint8_t err;
	if((err=rx_riff())!=0)
		return(err);
	
	// RIFF data
	uint8_t *riff=FIFO+RX_RIFF_OFS;
	
	// check RIFF tag
	if(memcmp_P((void*)riff,PSTR("RIFF"),4)!=0)
		return(4);
	riff+=8;
		
	// check WAVE tag
	if(memcmp_P((void*)riff,PSTR("WAVE"),4)!=0)
		return(5);
	riff+=4;
		
	// find "fmt " chunk
	uint8_t *fmt;
	while(1)
	{
		// is "fmt "?
		if(memcmp_P((void*)riff,PSTR("fmt "),4)==0)
		{
			// yeah - found
			riff+=4;

			// leave if out of range
			if(riff>FIFO+RX_RIFF_OFS+RX_RIFF_N-20)
				return(3);
			
			// past "fmt " chunk
			fmt=riff+*((uint16_t*)riff)+4;

			// check format
			riff+=4;
			if((*(uint16_t*)riff)!=1)
			{
				// not PCM
				return(6);
			}

			// channels count
			riff+=2;
			wave.chn=*riff;
			
			// sample rate
			riff+=2;
			wave.rate=*(uint16_t*)riff;
			I2S_config(wave.rate);

			// bit width
			riff+=10;
			wave.bit=*riff;
			
			break;
		}
		else
		{
			// nope - try again
			riff+=4;
			riff+=*((uint16_t*)riff)+4;
			
			// out of data
			if(riff>=FIFO+RX_RIFF_OFS+RX_RIFF_N-8)
				return(7);
				
		}
	}

	// find "data"
	riff=fmt;
	while(1)
	{
		// is "data"?
		if(memcmp_P((void*)riff,PSTR("data"),4)==0)
		{
			// yeah - found
			riff+=4;	

			// leave if not enough data
			if(riff>FIFO+RX_RIFF_OFS+RX_RIFF_N-4)
				return(3);

			// get wave data len
			wave.tord=*((uint32_t*)riff);

			// to data start
			riff+=4;

			// data to recieve (-=already_recieved_part)
			wave.tord-=(FIFOwr-riff);		
			
			// --- not the funny part - reformat recieved block to 16-bit stereo and align to first sample in FIFO ---
			fmt=FIFO;
			uint8_t be=0;
			if(wave.bit==16)
			{
				// 16-bit: endian swap
				if(wave.chn==2)
				{
					// stereo: just endian swap
					while(riff<FIFOwr)
					{
						if(be&1)
						{
							fmt[0]=*riff++;
							fmt+=2;
						}
						else
							fmt[1]=*riff++;
						be++;
					}
				}
				else
				{
					// mono: swap enadian, duplicate to stereo					
					while(riff<FIFOwr)
					{
						if(be&1)
						{
							fmt[0]=*riff;
							fmt[2]=*riff++;
							fmt+=4;
						}
						else
						{
							fmt[1]=*riff;
							fmt[3]=*riff++;
						}
						be++;
					}
				}				
			}
			else
			{
				// 8-bit: extend to 16-bit signed				
				if(wave.chn==2)
				{
					// stereo: convert to 16-bit signed
					while(riff<FIFOwr)
					{
						*fmt++=(*riff++)^0x80u;
						*fmt++=0x00;
					}
				}
				else
				{
					// mono: convert to 16-bit signed, duplicate
					while(riff<fmt)
					{
						*fmt++=(*riff)^0x80u;
						*fmt++=0x00;
						*fmt++=(*riff++)^0x80u;
						*fmt++=0x00;
					}
				}
			}

			// remember endian swap status
			wave.be=be;

			// write stire
			FIFOwr=fmt;
		
			// audio start
			I2S_set_read(FIFO);
			
			return(0);
		}
		else
		{
			// nope - try again
			riff+=4;
			riff+=*((uint16_t*)riff)+4;

			// out of data
			if(riff>=FIFO+RX_RIFF_OFS+RX_RIFF_N-8)
				return(8);
				
		}
	}
}



// ------ tx wave format string ------
void tx_format(void)
{
	char str[7];

	// sample rate
	utoa(wave.rate,str,10);
	tx_str(str);
	tx_str_P(PSTR(" Hz, "));

	// bit
	if(wave.bit==16)
		tx_str_P(PSTR("16-bit, "));
	else
		tx_str_P(PSTR("8-bit, "));

	// channels
	if(wave.chn==1)
		tx_str_P(PSTR("mono\n"));
	else
		tx_str_P(PSTR("stereo\n"));
}





//==========================================================================================================================
// Other
//==========================================================================================================================
// ------ tx error message ------
const char errmsg00[] PROGMEM="timeout!";
const char errmsg01[] PROGMEM="buffer overflow - probably disabled CTS handshake!";
const char errmsg02[] PROGMEM="failed!";
const char errmsg03[] PROGMEM="not a RIFF container!";
const char errmsg04[] PROGMEM="not a WAVE!";
const char errmsg05[] PROGMEM="not a PCM!";
const char errmsg06[] PROGMEM="\"fmt\" chunk not found!";
const char errmsg07[] PROGMEM="\"data\" chunk not found!";
const char* errmsg[8] PROGMEM={errmsg00,errmsg01,errmsg02,errmsg03,errmsg04,errmsg05,errmsg06,errmsg07};
void tx_error(uint8_t err)
{
	tx_str_P((PGM_P)pgm_read_word(&errmsg[err]));
	tx_str_P(PSTR("\n\n"));
}






//==========================================================================================================================
// Main
//==========================================================================================================================
int main(void)
{
	// directions (there are some unused signals beacause of unused peripherals on my breadboard)
	DDRA=0x00;
	DDRC=0x00;
	DDRB=1<<DWS|1<<DCK|1<<DDT|1<<RWR|1<<LRS|1<<LCK|1<<DCTS;
	DDRD=1<<IORD|1<<IOWR|1<<ROE|1<<RMA|1<<RHA|1<<BCO;

	// init signals
	PORTB=1<<RWR|0<<LCK|0<<DWS|1<<DCTS;
	PORTD=1<<IORD|1<<IOWR|1<<ROE|1<<RMA|1<<RHA|0<<BCO;
	
	// serial to full speed
	// baud=F_CPU/(8*(div+1))
	// div=0 for baud=2Mbd at F_CPU=16MHz
	// div=1 for baud=2Mbd at F_CPU=32MHz
	serial_config(0);

	// config timeout timer
	to_config();

	// allow IRQ
	sei();


	// --- MAIN LOOP ---
	tx_str_P(PSTR("\nPlayer initialization ... done.\n\n"));
	while(1)
	{
		tx_str_P(PSTR("\nWaiting for wave file ... "));
		
		// config player according to RIFF container
		uint8_t err;
		if((err=riff_config())!=0)
		{
			tx_error(err-1);

			// discard rest of data
			rx_discard();
			continue;
		}
		tx_format();
		tx_str_P(PSTR("Starting playback ... "));
		
		
		// set timeout
		to_time(750);
		to_run();

		// start DAC
		I2S_run();

		// CTS
		rx_cts_set();
	
	
		// --- main playback loop ---
		uint8_t *wr=FIFOwr;
		uint8_t be=wave.be;
		uint32_t tord=wave.tord;
		while(1)
		{
			// reset timeout
			to_reset();

			do{
				// get free space in FIFO
				int16_t spc;
				volatile uint8_t *rdtmp;
				ATOMIC_BLOCK(ATOMIC_FORCEON)
	    		{
					rdtmp=FIFOrd;
				}
				/*asm volatile("cli" ::: "memory");
				rdtmp=FIFOrd;
				asm volatile("sei" ::: "memory");*/
				
				spc=rdtmp-wr;
				if(spc<0)
					spc+=FIFOsize;

					
		
				// CTS shake
				if(spc<32)
					rx_cts_clr();
				if(spc>48)
					rx_cts_set();

				// leave if timeout
				if(to_test())
					break;

				// test for USART data
			}while(bit_is_clear(UCSR0A,RXC0));
			
			
			while(1)
			{
				// get byte
				uint8_t dat=UDR0;

				// select format
				if(wave.bit==16)
				{
					// 16-bit
					if(wave.chn==2)
					{
						// stereo
						if(be&1)
						{
							wr[0]=dat;
							wr+=2;
						}
						else
							wr[1]=dat;
						be++;
					}
					else
					{
						// mono: duplicate sample
						if(be&1)
						{
							wr[0]=dat;
							wr[2]=dat;
							wr+=4;
						}
						else
						{
							wr[1]=dat;
							wr[3]=dat;
						}
						be++;
					}
				}
				else
				{
					// 8-bit: extend to 16-bit signed
					*wr++=dat^0x80u;
					*wr++=0x00;
					if(wave.chn==1)
					{
						// mono: duplicate
						*wr++=dat^0x80u;
						*wr++=0x00;
					}
				}

				// reset pointer if overrun
				if(wr>=(unsigned)FIFOsize+FIFO)
					wr=FIFO;

				// repeat if something still there (this should happen 2 or 3 times at most)
				if(!--tord || bit_is_clear(UCSR0A,RXC0))
					break;
			}
			

			// check stream end or timeout
			if(to_test() || !tord)
			{
				// stream done or timeout

				// I2S silence output
				I2S_clear_sync();

				// I2S stop
				I2S_stop();

				// timeout stop
				to_stop();

				if(to_test())
					tx_str_P(PSTR("timeout!\n\n"));
				else
					tx_str_P(PSTR("done.\n\n"));

				// discard rest of stream data (sometimes there are additional RIFF chunks)
				rx_discard();

				break;
			}
		}
		FIFOwr=wr;
	}
}



